use crate::entity::helpers::check_assoc;
use crate::schema::{AssocMap, IDWrap};
use crate::{
    datum::{Datum, DatumList, DatumVisitor},
    entity::{Entity, EntityID, EntityPart, EntityPartList, EntityPartVisitor},
    schema::{EntityMap, IDMap},
};
use crate::{DBError, DBResult};
use std::hash::{Hash, Hasher};

#[derive(Hash)]
enum QueryType<'a> {
    Select(std::any::TypeId),
    SelectJoin(&'a str),
    Insert,
    InsertAssoc(&'a str),
}

fn query_hash<E: Entity>(ctx: &'static str, qtype: QueryType) -> u64 {
    let mut hasher = std::collections::hash_map::DefaultHasher::new();

    qtype.hash(&mut hasher);
    std::any::TypeId::of::<E>().hash(&mut hasher);
    ctx.hash(&mut hasher);

    hasher.finish()
}

pub(crate) fn select_assoc<E: Entity>(map: &AssocMap<E>) -> DBResult<Vec<E>> {
    let adata = map
        .data
        .as_ref()
        .ok_or(DBError::LogicError("Reading from AssocMap with no base ID"))?;
    // equivalent SQL:
    // SELECT
    //     `target_table_name`.*
    // FROM
    //     `assoc_table_name`
    // LEFT JOIN `target_table_name` ON `assoc_table_name`.`target` = `target_table_name`.`rowid`
    // WHERE `base` = base_rowid

    let ctx = adata.ctx;
    let base_name = adata.base_name;
    let target_name = E::entity_name();
    let part_name = adata.part_name;

    map.conn().with_prepared(
        query_hash::<E>(map.ctx(), QueryType::SelectJoin(part_name)),
        || {
            let assoc_name = format!("{base_name}_assoc_{part_name}_{target_name}");
            format!(
                "select `{ctx}_{target_name}`.* from `{ctx}_{assoc_name}` \
                    left join `{ctx}_{target_name}` on \
                        `{ctx}_{assoc_name}`.`target` = `{ctx}_{target_name}`.`ID` \
                    where `{ctx}_{assoc_name}`.`base` = ?"
            )
        },
        |stmt| {
            stmt.bind((1, adata.base_rowid))?;

            // now we grab the statement outputs
            let mut rows = vec![];
            while stmt.next()? == sqlite::State::Row {
                let datum_list = <E::Parts>::build_datum_list(&map.conn(), map.ctx(), stmt)?;
                rows.push(E::build(datum_list));
            }

            Ok(rows)
        },
    )
}

pub(crate) fn insert_assoc<E: Entity>(map: &AssocMap<E>, value: E) -> DBResult<()> {
    // we're doing two things:
    // - inserting the entity into the target table
    // - adding thw association row into the assoc table

    // so first, the target table
    let target_id = insert(map, value)?;

    let adata = map
        .data
        .as_ref()
        .ok_or(DBError::LogicError("Reading from AssocMap with no base ID"))?;

    // second, the assoc table
    map.conn().with_prepared(
        query_hash::<E>(map.ctx(), QueryType::InsertAssoc(adata.part_name)),
        || {
            let ctx = adata.ctx;
            let base_name = adata.base_name;
            let target_name = E::entity_name();
            let part_name = adata.part_name;
            let assoc_name = format!("{base_name}_assoc_{part_name}_{target_name}");
            format!("insert into `{ctx}_{assoc_name}` (`base`, `target`) values (?, ?)")
        },
        |stmt| {
            stmt.reset()?;
            stmt.bind((1, adata.base_rowid))?;
            stmt.bind((2, target_id.into_raw()))?;

            stmt.next()?;

            Ok(())
        },
    )
}

pub(crate) fn select_by<E: Entity, PL: EntityPartList>(
    map: &IDMap<E>,
    by: &PL::DatumList,
) -> DBResult<Vec<IDWrap<E>>> {
    struct HashDatumListTypes(std::collections::hash_map::DefaultHasher);
    impl DatumVisitor for HashDatumListTypes {
        fn visit<ED: Datum>(&mut self, _: &ED) {
            std::any::TypeId::of::<ED>().hash(&mut self.0);
        }
    }
    let mut ty = HashDatumListTypes(Default::default());
    by.accept(&mut ty);

    map.conn().with_prepared(
        query_hash::<E>(map.ctx(), QueryType::Select(std::any::TypeId::of::<PL>())),
        || {
            let mut conditions = String::new();
            struct BuildConditions<'a>(&'a mut String);
            impl<'a> EntityPartVisitor for BuildConditions<'a> {
                fn visit<EP: EntityPart>(&mut self) {
                    if self.0.len() > 0 {
                        self.0.push_str(" and ");
                    }
                    self.0.push_str("`");
                    self.0.push_str(EP::part_name());
                    self.0.push_str("`");
                    self.0.push_str(" = ?");
                }
            }
            PL::accept_part_visitor(&mut BuildConditions(&mut conditions));

            // CTX let table_name = format!("{}_{}", map.ctx(), E::entity_name());
            let table_name = format!("{}", E::entity_name());
            format!("select rowid, * from `{}` where {}", table_name, conditions)
        },
        |stmt| {
            struct BindDatum<'a, 'b>(&'a mut sqlite::Statement<'b>, usize);
            impl<'a, 'b> DatumVisitor for BindDatum<'a, 'b> {
                fn visit<ED: Datum>(&mut self, datum: &ED) {
                    datum.bind_to(self.0, self.1);
                    self.1 += 1
                }
            }
            // note that this indexing starts at 1
            by.accept(&mut BindDatum(stmt, 1));

            // now we grab the statement outputs
            let mut rows = vec![];
            while stmt.next()? == sqlite::State::Row {
                let datum_list = <E::Parts>::build_datum_list(&map.conn(), map.ctx(), stmt)?;
                rows.push(IDWrap::new(
                    <E::ID>::from_raw(stmt.read::<i64, _>(0)?),
                    E::build(datum_list),
                ));
            }

            stmt.reset()?;

            Ok(rows)
        },
    )
}

pub(crate) fn insert<E: Entity>(
    map: &impl EntityMap<ContainedEntity = E>,
    value: E,
) -> DBResult<E::ID> {
    map.conn().with_prepared(
        query_hash::<E>(map.ctx(), QueryType::Insert),
        || {
            // CTX let table_name = format!("{}_{}", map.ctx(), E::entity_name());
            let table_name = format!("{}", E::entity_name());

            let mut part_names = String::new();
            let mut placeholders = String::new();
            struct PartNameVisitor<'a>(&'a mut String, &'a mut String);
            impl<'a> EntityPartVisitor for PartNameVisitor<'a> {
                fn visit<EP: EntityPart>(&mut self) {
                    // if this is a set-association, then we don't actually want to do anything
                    // with it here; it doesn't have a column
                    if check_assoc::<EP>() {
                        return;
                    }

                    if self.0.len() != 0 {
                        self.0.push_str(", ");
                        self.1.push_str(", ");
                    }
                    self.0.push_str("`");
                    self.0.push_str(EP::part_name());
                    self.0.push_str("`");
                    self.1.push_str("?");
                }
            }

            E::accept_part_visitor(&mut PartNameVisitor(&mut part_names, &mut placeholders));

            format!(
                "insert into `{}` ({}) values ({}) returning `id`",
                table_name, part_names, placeholders
            )
        },
        |prepared| {
            prepared.reset()?;

            struct PartBinder<'a, 'b>(&'a mut sqlite::Statement<'b>, usize);
            impl<'a, 'b> EntityPartVisitor for PartBinder<'a, 'b> {
                fn visit_datum<EP: EntityPart>(&mut self, datum: &EP::Datum) {
                    // skip associations, as with the query preparation above
                    if check_assoc::<EP>() {
                        return;
                    }

                    datum.bind_to(self.0, self.1);
                    self.1 += 1;
                }
            }

            value.accept_part_visitor_ref(&mut PartBinder(prepared, 1));

            prepared.next()?;

            Ok(<E::ID>::from_raw(prepared.read::<i64, _>(0)?))
        },
    )
}
