use crate::{
    datum::Datum,
    db::{Connection, DBConnection},
    entity::{Entity, EntityPartList, EntityVisitor},
    query,
};
use crate::{DBError, DBResult};

mod build;
mod entity;
pub(crate) mod meta;
mod tests;

// ----------------------------------------------------------------------
// Specification types
// ----------------------------------------------------------------------

/// Trait for a type that represents a sqlite table that contains entities
pub(crate) trait EntityMap {
    type ContainedEntity: Entity;

    fn conn(&self) -> &DBConnection;
    fn ctx(&self) -> &'static str;
}

pub struct IDWrap<T: Entity> {
    id: T::ID,
    wrap: T,
}

impl<T: Entity> IDWrap<T> {
    pub fn new(id: T::ID, value: T) -> Self {
        Self { id, wrap: value }
    }

    pub fn id(&self) -> T::ID {
        self.id
    }

    pub fn as_ref(&self) -> &T {
        &self.wrap
    }
}

impl<T: Entity> AsRef<T> for IDWrap<T> {
    fn as_ref(&self) -> &T {
        &self.wrap
    }
}

impl<T: Entity> AsMut<T> for IDWrap<T> {
    fn as_mut(&mut self) -> &mut T {
        &mut self.wrap
    }
}

impl<T: Entity> std::ops::Deref for IDWrap<T> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        &self.wrap
    }
}

impl<T: Entity> std::ops::DerefMut for IDWrap<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.wrap
    }
}

/// Table with EntityID-based lookup.
pub struct IDMap<T: Entity> {
    conn: DBConnection,
    ctx: &'static str,
    _ghost: std::marker::PhantomData<T>,
}

impl<T: Entity> EntityMap for IDMap<T> {
    type ContainedEntity = T;
    fn conn(&self) -> &DBConnection {
        &self.conn
    }

    fn ctx(&self) -> &'static str {
        self.ctx
    }
}

impl<T: Entity> IDMap<T> {
    pub fn build(db: DBConnection, ctx: &'static str) -> Self {
        Self {
            conn: db,
            ctx,
            _ghost: std::marker::PhantomData,
        }
    }

    pub fn by_id(&self, _id: T::ID) -> DBResult<Option<T>> {
        // query::select_by
        todo!()
    }

    pub fn lookup_unique(
        &self,
        uniques: &<<T as Entity>::Uniques as EntityPartList>::DatumList,
    ) -> DBResult<Option<IDWrap<T>>> {
        query::select_by::<T, T::Uniques>(self, uniques).map(|mut v| {
            if v.len() > 0 {
                Some(v.remove(0))
            } else {
                None
            }
        })
    }

    pub fn insert(&self, value: T) -> DBResult<T::ID> {
        query::insert(self, value)
    }
}

pub struct Index<T: Entity, Key: Datum> {
    _conn: DBConnection,
    _ghost: std::marker::PhantomData<(T, Key)>,
}

pub struct AssocData {
    pub(crate) conn: DBConnection,
    pub(crate) ctx: &'static str,
    pub(crate) base_name: &'static str,
    pub(crate) part_name: &'static str,
    pub(crate) base_rowid: i64,
}

pub struct AssocMap<T: Entity> {
    pub(crate) data: Option<AssocData>,
    _ghost: std::marker::PhantomData<T>,
}

impl<T: Entity> std::fmt::Debug for AssocMap<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "AssocMap {{ id: {:?} }}",
            self.data.as_ref().map(|d| d.base_rowid)
        ))
    }
}

impl<T: Entity> AssocMap<T> {
    pub fn empty() -> Self {
        Self {
            data: None,
            _ghost: Default::default(),
        }
    }

    pub fn get_all(&self) -> DBResult<Vec<T>> {
        query::select_assoc(self)
    }

    pub fn insert(&self, value: T) -> DBResult<()> {
        query::insert_assoc(self, value)
    }
}

impl<T: Entity> EntityMap for AssocMap<T> {
    type ContainedEntity = T;

    fn conn(&self) -> &DBConnection {
        &self.data.as_ref().unwrap().conn
    }

    fn ctx(&self) -> &'static str {
        self.data.as_ref().unwrap().ctx
    }
}

impl<T: Entity> Datum for AssocMap<T> {
    fn sql_type() -> &'static str {
        unreachable!()
    }

    fn accept_entity_visitor(v: &mut impl EntityVisitor) {
        v.visit::<T>();
    }

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, _index: usize) {
        unreachable!()
    }

    fn build_from<'a>(
        adata: AssocData,
        _stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        // assuming that the stmt has the rowid as index 0
        Ok((
            Self {
                data: Some(adata),
                _ghost: Default::default(),
            },
            index,
        ))
    }
}

pub struct Serialized<T: serde::Serialize + serde::de::DeserializeOwned> {
    wrapped: T,
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + Default> Default for Serialized<T> {
    fn default() -> Self {
        Self {
            wrapped: T::default(),
        }
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned + std::fmt::Debug> std::fmt::Debug
    for Serialized<T>
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        <T as std::fmt::Debug>::fmt(&self.wrapped, f)
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned> From<T> for Serialized<T> {
    fn from(value: T) -> Self {
        Self { wrapped: value }
    }
}

impl<T: serde::Serialize + serde::de::DeserializeOwned> AsRef<T> for Serialized<T> {
    fn as_ref(&self) -> &T {
        &self.wrapped
    }
}

impl<T: 'static + serde::Serialize + serde::de::DeserializeOwned> Datum for Serialized<T> {
    fn sql_type() -> &'static str {
        "text"
    }

    fn bind_to<'a>(&self, stmt: &mut sqlite::Statement<'a>, index: usize) {
        <String as Datum>::bind_to(
            &serde_json::to_string(&self.wrapped).expect("couldn't serialize object into JSON"),
            stmt,
            index,
        )
    }

    fn build_from<'a>(
        adata: AssocData,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        let (s, idx) = <String as Datum>::build_from(adata, stmt, index)?;

        let d = serde_json::from_str::<T>(s.as_str()).map_err(|e| DBError::JSON(e))?;

        Ok((Self { wrapped: d }, idx))
    }
}

pub trait DatabaseItem {
    fn item_key() -> &'static str;
    fn dependency_keys() -> &'static [&'static str];

    fn is_index() -> bool {
        false
    }
    fn index_over() -> &'static str {
        unreachable!()
    }
    fn index_columns() -> &'static [&'static str] {
        unreachable!()
    }

    fn accept_entity_visitor(visitor: &mut impl EntityVisitor);
}

pub trait DatabaseItemVisitor {
    fn visit<DI: DatabaseItem>(&mut self)
    where
        Self: Sized;
}

pub trait DatabaseSpec {
    fn accept_entity_visitor(visitor: &mut impl EntityVisitor);
}

impl<T: Entity> DatabaseSpec for IDMap<T> {
    fn accept_entity_visitor(visitor: &mut impl EntityVisitor) {
        visitor.visit::<T>()
    }
}

pub trait Database {
    fn open_path<U: AsRef<str>>(uri: U) -> Result<Self, DBError>
    where
        Self: Sized,
    {
        let conn = Connection::open(uri)?;
        let schema = build::collect_from_database::<Self>();
        match schema.check(conn.clone()) {
            // schema checks out
            Some(true) => {}
            // schema doesn't match
            Some(false) => Err(DBError::IncompatibleSchema)?,
            // no schema found
            None => {
                schema.create(conn.clone())?;
            }
        }

        Ok(Self::build(conn))
    }

    #[doc(hidden)]
    fn build(conn: DBConnection) -> Self
    where
        Self: Sized;

    fn accept_item_visitor(visitor: &mut impl DatabaseItemVisitor)
    where
        Self: Sized;
}
